/*
  Gabriel Massaki Wakano Bezerra
  Emely Pujolli da Silva
*/

#include <SOIL.h>
#include <iostream>
#include <algorithm>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>

#include <GL/freeglut.h>

#include <glm/glm.hpp> 
#include <glm/gtc/matrix_transform.hpp>
#include <glm/gtc/type_ptr.hpp>

#include "LoadShaders.h"

#define MAX_PLANTS 5		// Maximum number of plants for plant forests
#define GRID_RESOLVE 65		// Size of the surface grid

using namespace std;

/******************************************************************************
Data structures section
*******************************************************************************/
// Tree structure used to hold the plant model
struct PlantNode {
  char type;         // Node type, a=stem, b=stem, c=leaf, d=flower    
  GLfloat z_ang;     // Rotation of this node w.r.t. parent's z-axis   - Rotates around the current stem direction
  GLfloat x_ang;     // Rotation of this node w.r.t. parent's x-axis   - Rotates away from current stem direction
  struct PlantNode *left;     // Left child for this node (NULL means the node is a terminal)
  struct PlantNode *right;    // Right child for this node (NULL means the node is a terminal)
};

/******************************************************************************
Global data
******************************************************************************/
float aspect;

GLuint shaders;
GLuint vao;
GLuint vertexbuffer;
GLuint colorbuffer;
GLuint normalbuffer;
GLuint prjMatrix;
GLuint mvMatrix;
GLuint texture;

double GroundXYZ[GRID_RESOLVE][GRID_RESOLVE][3];	// Array to store ground surface points as (x,y,z) points
double GroundNormals[GRID_RESOLVE][GRID_RESOLVE][3];   // Array to hold normal vectors at vertices (will use normal to triangle
// defined by the vertices at [i][j], [i+1][j], and [i][j+1]

// Transition probabilities for the L-system specification
GLfloat Paab;
GLfloat Paac;
GLfloat Paad;
GLfloat Pacd;
GLfloat Pba;
GLfloat Pbc;
GLfloat Pbd;

/* ArcBall */
int lastX = 0, lastY = 0, curX = 0, curY = 0;
int arcball_on = false;
glm::mat4 tc_matrix;

// These affect the shape of the plants, and control the number of plants in the forest
GLfloat Z_angle;      // Max angle around local Z-axis for each level
GLfloat X_angle;      // Max angle around local X-axis at each branching point
GLint n_levels;       // Number of levels of branching in the plant

GLfloat vertexbuffer_data[1000000]; // Grid + Plants
GLfloat normalbuffer_data[1000000];
GLfloat colorbuffer_data[1000000];

// Global indeces
int indexNormal, indexVertex, indexColor;
// Numbers of triangles
int numTri;

/******************************************************************************
  Function Prototypes
*******************************************************************************/
// GLUT
void Initialize(const char *title);
void Display(void);
void Finalize(void);
void Reshape(int width, int height);

// L-system
struct PlantNode *makePlant (void);
void generateRecursivePlant (struct PlantNode *p, int level);
void freePlant(struct PlantNode *p);
void printPlant(struct PlantNode *p);
void printNodeRecursive(struct PlantNode *p, int lev, int tgt);
void toBufferPlant(struct PlantNode *p, glm::mat4 matrix);
void stemSection(glm::mat4 matrix);
void leafSection(glm::mat4 matrix);
void flowerSection(glm::mat4 matrix);

// Surface generation
void makeSurfaceGrid(void);
void toBuffers(void);
void computeNormal(GLfloat *vx, GLfloat *vy, GLfloat *vz, GLfloat wx, GLfloat wy, GLfloat wz);

// Auxiliar
void computeNormal(GLfloat *vx, GLfloat *vy, GLfloat *vz, GLfloat wx, GLfloat wy, GLfloat wz);
double noise(double x, double y);
double rand48(void);
void toNormalBuffer(GLfloat a, GLfloat b, GLfloat c);
void toVertexBuffer(GLfloat a, GLfloat b, GLfloat c);
void toColorBuffer(GLfloat a, GLfloat b, GLfloat c);

/******************************************************************************
Code
*******************************************************************************/

/* ArcBall */
void onMouse(int button, int state, int x, int y) {
  if (button == GLUT_LEFT_BUTTON && state == GLUT_DOWN) {
    arcball_on = true;
    lastX = curX = x;
    lastY = curX = y;
  }
  else {
    arcball_on = false;
  }
}

/* ArcBall */
void onMotion(int x, int y) {
  if (arcball_on) {  // if left button is pressed
    curX = x;
    curY = y;
  }
}

/**
 * Get a normalized vector from the center of the virtual ball O to a
 * point P on the virtual ball surface, such that P is aligned on
 * screen's (X,Y) coordinates.  If (X,Y) is too far away from the
 * sphere, return the nearest point on the virtual ball surface.
 */
glm::vec3 get_arcball_vector(int x, int y) {
  int screen_width = glutGet(GLUT_WINDOW_WIDTH);
  int screen_height = glutGet(GLUT_WINDOW_HEIGHT);
  glm::vec3 P = glm::vec3(1.0 * x / screen_width * 2 - 1.0,
			  1.0 * y / screen_height * 2 - 1.0,
			  0);
  P.y = -P.y;
  float OP_squared = P.x * P.x + P.y * P.y;
  if (OP_squared <= 1 * 1)
    P.z = sqrt(1 * 1 - OP_squared);  // Pythagore
  else
    P = glm::normalize(P);  // nearest point
  return P;
}


// Compute the normal to a surface given two vectors known to be on that surface
void computeNormal(GLfloat *vx, GLfloat *vy, GLfloat *vz, GLfloat wx, GLfloat wy, GLfloat wz) {
  GLfloat len;
  GLfloat nx, ny, nz;

  len = sqrt(((*vx) * (*vx)) + ((*vy) * (*vy)) + ((*vz) * (*vz)));
  *(vx) = (*vx) / len;
  *(vy) = (*vy) / len;
  *(vz) = (*vz) / len;

  len = sqrt((wx * wx) + (wy * wy) + (wz * wz));
  wx /= len;
  wy /= len;
  wz /= len;

  nx = ((*vy) * wz) - (wy * (*vz));
  ny = (wx * (*vz)) - ((*vx) * wz);
  nz = ((*vx) * wy) - (wx * (*vy));
  len = sqrt(((nx) * (nx)) + ((ny) * (ny)) + ((nz) * (nz)));
  *(vx) = nx / len;
  *(vy) = ny / len;
  *(vz) = nz / len;
}

// http://www.dreamincode.net/forums/topic/66480-perlin-noise/
// Returns a value between -0.25 and 0.25, used to generate noise
double noise(double x, double y) {
  int n, nn;
  n = (int)x + (int)y * 57;
  n = (n << 13) ^ n;
  nn = (n * (n * n * 60493 + 19990303) + 1376312589) & 0x7fffffff;
  return (double)(1.0 - ((double)nn / 1073741824.0)) / 4;
}

double rand48() {
  int n = rand() % 100;
  return (double) (n / 100.0);
}

void toNormalBuffer(GLfloat a, GLfloat b, GLfloat c) {
  normalbuffer_data[indexNormal] = a;
  normalbuffer_data[indexNormal + 1] = b;
  normalbuffer_data[indexNormal + 2] = c;
  indexNormal = indexNormal + 3;
  numTri++;
}

void toVertexBuffer(GLfloat a, GLfloat b, GLfloat c) {
  vertexbuffer_data[indexVertex] = a;
  vertexbuffer_data[indexVertex + 1] = b;
  vertexbuffer_data[indexVertex + 2] = c;
  indexVertex = indexVertex + 3;
  numTri++;
}

void toColorBuffer(GLfloat a, GLfloat b, GLfloat c) {
  colorbuffer_data[indexColor] = a;
  colorbuffer_data[indexColor + 1] = b;
  colorbuffer_data[indexColor + 2] = c;
  indexColor = indexColor + 3;
  numTri++;
}

// Fill the buffers with the data
void toBuffers(void) {
	
  /*

    1--------3--------3--------3--------2
    |      - |      - |      - |      - |
    |    -   |    -   |    -   |    -   |
    |  -     |  -     |  -     |  -     |
    |-       |-       |-       |-       |
    3--------6--------6--------6--------3
    |      - |      - |      - |      - |
    |    -   |    -   |    -   |    -   |
    |  -     |  -     |  -     |  -     |
    |-       |-       |-       |-       |
    3--------6--------6--------6--------3
    |      - |      - |      - |      - |
    |    -   |    -   |    -   |    -   |
    |  -     |  -     |  -     |  -     |
    |-       |-       |-       |-       |
    3--------6--------6--------6--------3
    |      - |      - |      - |      - |
    |    -   |    -   |    -   |    -   |
    |  -     |  -     |  -     |  -     |
    |-       |-       |-       |-       |
    2--------3--------3--------3--------1
	
    #s = points(x, y)
    - = triangle lines

    The numbers represent how many triangles the vertix is part of

  */
		
  // The conditions cover all possibilities of vertices and connect the triangles in the right way
  for (int i = 0; i < GRID_RESOLVE; i++) {
    for (int j = 0; j < GRID_RESOLVE; j++) {
      if (i + 1 < GRID_RESOLVE && j + 1 < GRID_RESOLVE) {
  toNormalBuffer(GroundNormals[i][j][0], GroundNormals[i][j][1], GroundNormals[i][j][2]);
	toNormalBuffer(GroundXYZ[i + 1][j][0], GroundXYZ[i + 1][j][1], GroundXYZ[i + 1][j][2]);
	toNormalBuffer(GroundXYZ[i][j + 1][0], GroundXYZ[i][j + 1][1], GroundXYZ[i][j + 1][2]);
	toVertexBuffer(GroundXYZ[i][j][0], GroundXYZ[i][j][1], GroundXYZ[i][j][2] - 10);
	toVertexBuffer(GroundXYZ[i + 1][j][0], GroundXYZ[i + 1][j][1], GroundXYZ[i + 1][j][2] - 10);
	toVertexBuffer(GroundXYZ[i][j + 1][0], GroundXYZ[i][j + 1][1], GroundXYZ[i][j + 1][2] - 10);
	toColorBuffer(0.5f, 0.5f, 0.5f);
	toColorBuffer(0.5f, 0.5f, 0.5f);
	toColorBuffer(0.5f, 0.5f, 0.5f);
      }
      if (i - 1 >= 0 && j + 1 < GRID_RESOLVE) {
	toNormalBuffer(GroundXYZ[i][j][0], GroundXYZ[i][j][1], GroundXYZ[i][j][2]);
	toNormalBuffer(GroundXYZ[i - 1][j][0], GroundXYZ[i - 1][j][1], GroundXYZ[i - 1][j][2]);
	toNormalBuffer(GroundXYZ[i][j + 1][0], GroundXYZ[i][j + 1][1], GroundXYZ[i][j + 1][2]);
	toVertexBuffer(GroundXYZ[i][j][0], GroundXYZ[i][j][1], GroundXYZ[i][j][2] - 10);
	toVertexBuffer(GroundXYZ[i - 1][j][0], GroundXYZ[i - 1][j][1], GroundXYZ[i - 1][j][2] - 10);
	toVertexBuffer(GroundXYZ[i][j + 1][0], GroundXYZ[i][j + 1][1], GroundXYZ[i][j + 1][2] - 10);
	toColorBuffer(0.5f, 0.5f, 0.5f);
	toColorBuffer(0.5f, 0.5f, 0.5f);
	toColorBuffer(0.5f, 0.5f, 0.5f);

	toNormalBuffer(GroundXYZ[i][j][0], GroundXYZ[i][j][1], GroundXYZ[i][j][2]);
	toNormalBuffer(GroundXYZ[i - 1][j + 1][0], GroundXYZ[i - 1][j + 1][1], GroundXYZ[i - 1][j + 1][2]);
	toNormalBuffer(GroundXYZ[i - 1][j][0], GroundXYZ[i - 1][j][1], GroundXYZ[i - 1][j][2]);
	toVertexBuffer(GroundXYZ[i][j][0], GroundXYZ[i][j][1], GroundXYZ[i][j][2] - 10);
	toVertexBuffer(GroundXYZ[i - 1][j + 1][0], GroundXYZ[i - 1][j + 1][1], GroundXYZ[i - 1][j + 1][2] - 10);
	toVertexBuffer(GroundXYZ[i - 1][j][0], GroundXYZ[i - 1][j][1], GroundXYZ[i - 1][j][2] - 10);
	toColorBuffer(0.5f, 0.5f, 0.5f);
	toColorBuffer(0.5f, 0.5f, 0.5f);
	toColorBuffer(0.5f, 0.5f, 0.5f);
      }
      if (i - 1 >= 0 && j - 1 >= 0) {
	toNormalBuffer(GroundXYZ[i][j][0], GroundXYZ[i][j][1], GroundXYZ[i][j][2]);
	toNormalBuffer(GroundXYZ[i - 1][j][0], GroundXYZ[i - 1][j][1], GroundXYZ[i - 1][j][2]);
	toNormalBuffer(GroundXYZ[i][j - 1][0], GroundXYZ[i][j - 1][1], GroundXYZ[i][j - 1][2]);
	toVertexBuffer(GroundXYZ[i][j][0], GroundXYZ[i][j][1], GroundXYZ[i][j][2] - 10);
	toVertexBuffer(GroundXYZ[i - 1][j][0], GroundXYZ[i - 1][j][1], GroundXYZ[i - 1][j][2] - 10);
	toVertexBuffer(GroundXYZ[i][j - 1][0], GroundXYZ[i][j - 1][1], GroundXYZ[i][j - 1][2] - 10);
	toColorBuffer(0.5f, 0.5f, 0.5f);
	toColorBuffer(0.5f, 0.5f, 0.5f);
	toColorBuffer(0.5f, 0.5f, 0.5f);
      }
      if (i + 1 < GRID_RESOLVE && j - 1 >= 0) {
	toNormalBuffer(GroundXYZ[i][j][0], GroundXYZ[i][j][1], GroundXYZ[i][j][2]);
	toNormalBuffer(GroundXYZ[i][j - 1][0], GroundXYZ[i][j - 1][1], GroundXYZ[i][j - 1][2]);
	toNormalBuffer(GroundXYZ[i + 1][j - 1][0], GroundXYZ[i + 1][j - 1][1], GroundXYZ[i + 1][j - 1][2]);
	toVertexBuffer(GroundXYZ[i][j][0], GroundXYZ[i][j][1], GroundXYZ[i][j][2] - 10);
	toVertexBuffer(GroundXYZ[i][j - 1][0], GroundXYZ[i][j - 1][1], GroundXYZ[i][j - 1][2] - 10);
	toVertexBuffer(GroundXYZ[i + 1][j - 1][0], GroundXYZ[i + 1][j - 1][1], GroundXYZ[i + 1][j - 1][2] - 10);
	toColorBuffer(0.5f, 0.5f, 0.5f);
	toColorBuffer(0.5f, 0.5f, 0.5f);
	toColorBuffer(0.5f, 0.5f, 0.5f);

	toNormalBuffer(GroundXYZ[i][j][0], GroundXYZ[i][j][1], GroundXYZ[i][j][2]);
	toNormalBuffer(GroundXYZ[i + 1][j - 1][0], GroundXYZ[i + 1][j - 1][1], GroundXYZ[i + 1][j - 1][2]);
	toNormalBuffer(GroundXYZ[i + 1][j][0], GroundXYZ[i + 1][j][1], GroundXYZ[i + 1][j][2]);
	toVertexBuffer(GroundXYZ[i][j][0], GroundXYZ[i][j][1], GroundXYZ[i][j][2] - 10);
	toVertexBuffer(GroundXYZ[i + 1][j - 1][0], GroundXYZ[i + 1][j - 1][1], GroundXYZ[i + 1][j - 1][2] - 10);
	toVertexBuffer(GroundXYZ[i + 1][j][0], GroundXYZ[i + 1][j][1], GroundXYZ[i + 1][j][2] - 10);
	toColorBuffer(0.5f, 0.5f, 0.5f);
	toColorBuffer(0.5f, 0.5f, 0.5f);
	toColorBuffer(0.5f, 0.5f, 0.5f);
      }
    }
  }
}

// Midpoint displacement algorithm with noise
void generateTerrain(int x1, int y1, int x2, int y2, int x3, int y3, int x4, int y4, GLfloat side) {

  /*

    1-----------------------------------2
    |                 |                 |
    |                 |                 |
    |                 |                 |
    |                 |                 |
    |                 |                 |
    |                 |                 |
    |                 |                 |
    |                 |                 |
    |                 |                 |
    |----------------mid----------------|
    |                 |                 |
    |                 |                 |
    |                 |                 |
    |                 |                 |
    |                 |                 |
    |                 |                 |
    |                 |                 |
    |                 |                 |
    |                 |                 |
    3-----------------------------------4


    #s = points(x, y)
    - = triangle lines

    The numbers represent how many triangles the vertix is part of

  */

  // If you can not divide it means the points are next to each other and there is no point between them
  if (x2 - x1 < 2)
    return;

  // Calculating the top side value
  GroundXYZ[(x1 + x2) / 2][(y1 + y2) / 2][0] = (-side*.5) + (((x1 + x2) / 2) * (side / GRID_RESOLVE));
  GroundXYZ[(x1 + x2) / 2][(y1 + y2) / 2][1] = (-side*.5) + (((y1 + y2) / 2) * (side / GRID_RESOLVE));
  GroundXYZ[(x1 + x2) / 2][(y1 + y2) / 2][2] = (GroundXYZ[x1][y1][2] + GroundXYZ[x2][y2][2]) / 2;
  GroundXYZ[(x1 + x2) / 2][(y1 + y2) / 2][2] += noise((x1 + x2) / 2, (y1 + y2) / 2) * GroundXYZ[(x1 + x2) / 2][(y1 + y2) / 2][2];
  // Calculating the left side value
  GroundXYZ[(x1 + x3) / 2][(y1 + y3) / 2][0] = (-side*.5) + (((x1 + x3) / 2) * (side / GRID_RESOLVE));
  GroundXYZ[(x1 + x3) / 2][(y1 + y3) / 2][1] = (-side*.5) + (((y1 + y3) / 2) * (side / GRID_RESOLVE));
  GroundXYZ[(x1 + x3) / 2][(y1 + y3) / 2][2] = (GroundXYZ[x1][y1][2] + GroundXYZ[x3][y3][2]) / 2;
  GroundXYZ[(x1 + x3) / 2][(y1 + y3) / 2][2] += noise((x1 + x3) / 2, (y1 + y3) / 2) * GroundXYZ[(x1 + x3) / 2][(y1 + y3) / 2][2];
  // Calculating the right side value
  GroundXYZ[(x2 + x4) / 2][(y2 + y4) / 2][0] = (-side*.5) + (((x2 + x4) / 2) * (side / GRID_RESOLVE));
  GroundXYZ[(x2 + x4) / 2][(y2 + y4) / 2][1] = (-side*.5) + (((y2 + y4) / 2) * (side / GRID_RESOLVE));
  GroundXYZ[(x2 + x4) / 2][(y2 + y4) / 2][2] = (GroundXYZ[x2][y2][2] + GroundXYZ[x4][y4][2]) / 2;
  GroundXYZ[(x2 + x4) / 2][(y2 + y4) / 2][2] += noise((x2 + x4) / 2, (y2 + y4) / 2) * GroundXYZ[(x2 + x4) / 2][(y2 + y4) / 2][2];
  // Calculating the bottom side value
  GroundXYZ[(x3 + x4) / 2][(y3 + y4) / 2][0] = (-side*.5) + (((x3 + x4) / 2) * (side / GRID_RESOLVE));
  GroundXYZ[(x3 + x4) / 2][(y3 + y4) / 2][1] = (-side*.5) + (((y3 + y4) / 2) * (side / GRID_RESOLVE));
  GroundXYZ[(x3 + x4) / 2][(y3 + y4) / 2][2] = (GroundXYZ[x3][y3][2] + GroundXYZ[x4][y4][2]) / 2;
  GroundXYZ[(x3 + x4) / 2][(y3 + y4) / 2][2] += noise((x3 + x4) / 2, (y3 + y4) / 2) * GroundXYZ[(x3 + x4) / 2][(y3 + y4) / 2][2];
  // Calculating the mid value
  GroundXYZ[(x1 + x2 + x3 + x4) / 4][(y1 + y2 + y3 + y4) / 4][0] = (-side*.5) + (((x1 + x2 + x3 + x4) / 4) * (side / GRID_RESOLVE));
  GroundXYZ[(x1 + x2 + x3 + x4) / 4][(y1 + y2 + y3 + y4) / 4][1] = (-side*.5) + (((y1 + y2 + y3 + y4) / 4) * (side / GRID_RESOLVE));
  GroundXYZ[(x1 + x2 + x3 + x4) / 4][(y1 + y2 + y3 + y4) / 4][2] = (GroundXYZ[x1][y1][2] + GroundXYZ[x2][y2][2] +	GroundXYZ[x3][y3][2] + GroundXYZ[x4][y4][2]) / 4;
  GroundXYZ[(x1 + x2 + x3 + x4) / 4][(y1 + y2 + y3 + y4) / 4][2] += noise((x1 + x2 + x3 + x4) / 4, (y1 + y2 + y3 + y4) / 4) * GroundXYZ[(x1 + x2 + x3 + x4) / 4][(y1 + y2 + y3 + y4) / 4][2];
  // Calling the function for all the squares
  generateTerrain(x1, y1, (x1 + x2) / 2, (y1 + y2) / 2, (x1 + x3) / 2, (y1 + y3) / 2,	(x1 + x2 + x3 + x4) / 4, (y1 + y2 + y3 + y4) / 4, side);
  generateTerrain((x1 + x2) / 2, (y1 + y2) / 2, x2, y2, (x1 + x2 + x3 + x4) / 4, (y1 + y2 + y3 + y4) / 4,	(x2 + x4) / 2, (y2 + y4) / 2, side);
  generateTerrain((x1 + x3) / 2, (y1 + y3) / 2, (x1 + x2 + x3 + x4) / 4, (y1 + y2 + y3 + y4) / 4,	x3, y3, (x3 + x4) / 2, (y3 + y4) / 2, side);
  generateTerrain((x1 + x2 + x3 + x4) / 4, (y1 + y2 + y3 + y4) / 4, (x2 + x4) / 2, (y2 + y4) / 2,	(x3 + x4) / 2, (y3 + y4) / 2, x4, y4, side);
}

// Generate a surface to place the plants on
void makeSurfaceGrid(void) {
  GLfloat side;
  GLfloat vx, vy, vz, wx, wy, wz;
  GLfloat auxX, auxY, auxZ, len;
  int counter; // Stores the number of triangles the point is part of
  // Assign surface heights
  side = 15;				// Width of the surface - X and Y coordinates
  // will have values in [-side/2, side/2]

  srand(time(NULL));

  // Creates the initial 4 points
  GroundXYZ[0][0][0] = -side * 0.5;
  GroundXYZ[0][0][1] = -side * 0.5;
  GroundXYZ[0][0][2] = rand() % 4;

  GroundXYZ[GRID_RESOLVE - 1][0][0] = side * 0.5;
  GroundXYZ[GRID_RESOLVE - 1][0][1] = -side * 0.5;
  GroundXYZ[GRID_RESOLVE - 1][0][2] = rand() % 4;

  GroundXYZ[0][GRID_RESOLVE - 1][0] = -side * 0.5;
  GroundXYZ[0][GRID_RESOLVE - 1][1] = side * 0.5;
  GroundXYZ[0][GRID_RESOLVE - 1][2] = rand() % 4;

  GroundXYZ[GRID_RESOLVE - 1][GRID_RESOLVE - 1][0] = side * 0.5;
  GroundXYZ[GRID_RESOLVE - 1][GRID_RESOLVE - 1][1] = side * 0.5;
  GroundXYZ[GRID_RESOLVE - 1][GRID_RESOLVE - 1][2] = rand() % 4;

  // Generates the terrain
  generateTerrain(0, 0, GRID_RESOLVE - 1, 0, 0, GRID_RESOLVE - 1, GRID_RESOLVE - 1, GRID_RESOLVE - 1, side);

  // Calculates the normal vetor for each point, following the triagles shown at the toBuffers function
  // Because one point can be part of more than one triangle, we sum the normals the normalize afterwards
  // Obtain two vectors on the surface the point at GroundXYZ[i][j][] is located
  for (int i = 0; i < GRID_RESOLVE; i++) {
    for (int j = 0; j < GRID_RESOLVE; j++) {
      auxX = 0;
      auxY = 0;
      auxZ = 0;
      counter = 1;
      if (i + 1 < GRID_RESOLVE && j + 1 < GRID_RESOLVE) {
	vx = GroundXYZ[i + 1][j][0] - GroundXYZ[i][j][0];
	vy = GroundXYZ[i + 1][j][1] - GroundXYZ[i][j][1];
	vz = GroundXYZ[i + 1][j][2] - GroundXYZ[i][j][2];
	wx = GroundXYZ[i][j + 1][0] - GroundXYZ[i][j][0];
	wy = GroundXYZ[i][j + 1][1] - GroundXYZ[i][j][1];
	wz = GroundXYZ[i][j + 1][2] - GroundXYZ[i][j][2];
	// compute the normal
	computeNormal(&vx, &vy, &vz, wx, wy, wz);
	auxX += vx;
	auxY += vy;
	auxZ += vz;
	counter++;
      }
      if (i - 1 >= 0 && j + 1 < GRID_RESOLVE) {
	vx = GroundXYZ[i - 1][j + 1][0] - GroundXYZ[i][j][0];
	vy = GroundXYZ[i - 1][j + 1][1] - GroundXYZ[i][j][1];
	vz = GroundXYZ[i - 1][j + 1][2] - GroundXYZ[i][j][2];
	wx = GroundXYZ[i][j + 1][0] - GroundXYZ[i][j][0];
	wy = GroundXYZ[i][j + 1][1] - GroundXYZ[i][j][1];
	wz = GroundXYZ[i][j + 1][2] - GroundXYZ[i][j][2];
	// compute the normal
	computeNormal(&vx, &vy, &vz, wx, wy, wz);
	auxX += vx;
	auxY += vy;
	auxZ += vz;
	vx = GroundXYZ[i - 1][j + 1][0] - GroundXYZ[i][j][0];
	vy = GroundXYZ[i - 1][j + 1][1] - GroundXYZ[i][j][1];
	vz = GroundXYZ[i - 1][j + 1][2] - GroundXYZ[i][j][2];
	wx = GroundXYZ[i - 1][j][0] - GroundXYZ[i][j][0];
	wy = GroundXYZ[i - 1][j][1] - GroundXYZ[i][j][1];
	wz = GroundXYZ[i - 1][j][2] - GroundXYZ[i][j][2];
	// compute the normal
	computeNormal(&vx, &vy, &vz, wx, wy, wz);
	auxX += vx;
	auxY += vy;
	auxZ += vz;
	counter += 2;
      }
      if (i - 1 >= 0 && j - 1 >= 0) {
	vx = GroundXYZ[i - 1][j][0] - GroundXYZ[i][j][0];
	vy = GroundXYZ[i - 1][j][1] - GroundXYZ[i][j][1];
	vz = GroundXYZ[i - 1][j][2] - GroundXYZ[i][j][2];
	wx = GroundXYZ[i][j - 1][0] - GroundXYZ[i][j][0];
	wy = GroundXYZ[i][j - 1][1] - GroundXYZ[i][j][1];
	wz = GroundXYZ[i][j - 1][2] - GroundXYZ[i][j][2];
	// compute the normal
	computeNormal(&vx, &vy, &vz, wx, wy, wz);
	auxX += vx;
	auxY += vy;
	auxZ += vz;
	counter++;
      }
      if (i + 1 < GRID_RESOLVE && j - 1 >= 0) {
	vx = GroundXYZ[i][j - 1][0] - GroundXYZ[i][j][0];
	vy = GroundXYZ[i][j - 1][1] - GroundXYZ[i][j][1];
	vz = GroundXYZ[i][j - 1][2] - GroundXYZ[i][j][2];
	wx = GroundXYZ[i + 1][j - 1][0] - GroundXYZ[i][j][0];
	wy = GroundXYZ[i + 1][j - 1][1] - GroundXYZ[i][j][1];
	wz = GroundXYZ[i + 1][j - 1][2] - GroundXYZ[i][j][2];
	// compute the normal
	computeNormal(&vx, &vy, &vz, wx, wy, wz);
	auxX += vx;
	auxY += vy;
	auxZ += vz;
	vx = GroundXYZ[i + 1][j - 1][0] - GroundXYZ[i][j][0];
	vy = GroundXYZ[i + 1][j - 1][1] - GroundXYZ[i][j][1];
	vz = GroundXYZ[i + 1][j - 1][2] - GroundXYZ[i][j][2];
	wx = GroundXYZ[i + 1][j][0] - GroundXYZ[i][j][0];
	wy = GroundXYZ[i + 1][j][1] - GroundXYZ[i][j][1];
	wz = GroundXYZ[i + 1][j][2] - GroundXYZ[i][j][2];
	// compute the normal
	computeNormal(&vx, &vy, &vz, wx, wy, wz);
	auxX += vx;
	auxY += vy;
	auxZ += vz;
	counter += 2;
      }
      auxX = auxX / counter;
      auxY = auxY / counter;
      auxZ = auxZ / counter;
      len = sqrt(((auxX) * (auxX)) + ((auxY) * (auxY)) + ((auxZ) * (auxZ)));
      // And store it...
      GroundNormals[i][j][0] = auxX / len;
      GroundNormals[i][j][1] = auxY / len;
      GroundNormals[i][j][2] = auxZ / len;
    }
  }
}

// Recursive rendering function for the plant. Renders the
// section specified by p after performing the transformations
// required to give this part the correct orientation and
// position.
void toBufferPlant(struct PlantNode *p, glm::mat4 matrix) {

  if (p == NULL) // Avoid crash if called with empty node
    return;		

  // Rotates and scales
  matrix = matrix * glm::rotate(glm::mat4(1.0), glm::degrees(p->x_ang), glm::vec3(1.0f, 0.0f, 0.0f));
  matrix = matrix * glm::rotate(glm::mat4(1.0), glm::degrees(p->z_ang), glm::vec3(0.0f, 0.0f, 1.0f));
	
  if (p->type == 'a' || p->type == 'b') {
    // Draws the stem
    stemSection(matrix);
    matrix = matrix * glm::translate(glm::mat4(1.0), glm::vec3(0.0f, 0.0f, 1.0f));
    // Draws a flower if it is the end
    if (p->right == NULL && p->left == NULL) {
      flowerSection(matrix);
    }
    toBufferPlant(p->right, matrix);
    toBufferPlant(p->left, matrix);
  }
  else if (p->type == 'c') {
    // Draws a leaf
    leafSection(matrix);
  }
  else if (p->type == 'd') {
    // Draws a flower
    flowerSection(matrix);
  }
}

void stemSection(glm::mat4 matrix) {
  // Draws a single stem section, along the current local Z axis
  glm::vec4 p;
  GLfloat vx, vy, vz, wx, wy, wz;
	
  // Cube
  // 1
  // 1.1
  // Normal
  vx = 0.05 - (-0.05);
  vy = 0.05 - 0.05;
  vz = 0.0 - 0.0;
  wx = -0.05 - (-0.05);
  wy = 0.05 - 0.05;
  wz = 0.10 - 0.0;
  computeNormal(&vx, &vy, &vz, wx, wy, wz);
  p = glm::vec4(vx, vy, vz, 1.0f);
  p = matrix * p;
  toNormalBuffer(p[0], p[1], p[2]);
  vx = -0.05 - 0.05;
  vy = 0.05 - 0.05;
  vz = 0.0 - 0.0;
  wx = -0.05 - 0.05;
  wy = 0.05 - 0.05;
  wz = 0.10 - 0.0;
  computeNormal(&vx, &vy, &vz, wx, wy, wz);
  p = glm::vec4(vx, vy, vz, 1.0f);
  p = matrix * p;
  toNormalBuffer(p[0], p[1], p[2]);
  vx = -0.05 - (-0.05);
  vy = 0.05 - 0.05;
  vz = 0.0 - 0.10;
  wx = 0.05 - (-0.05);
  wy = 0.05 - 0.05;
  wz = 0.0 - 0.10;
  computeNormal(&vx, &vy, &vz, wx, wy, wz);
  p = glm::vec4(vx, vy, vz, 1.0f);
  p = matrix * p;
  toNormalBuffer(p[0], p[1], p[2]);
  // Vertex
  p = glm::vec4(-0.05f, 0.05f, 0.0f, 1.0f);
  p = matrix * p;
  toVertexBuffer(p[0], p[1], p[2]);
  p = glm::vec4(0.05f, 0.05f, 0.0f, 1.0f);
  p = matrix * p;
  toVertexBuffer(p[0], p[1], p[2]);
  p = glm::vec4(-0.05f, 0.05f, 1.0f, 1.0f);
  p = matrix * p;
  toVertexBuffer(p[0], p[1], p[2]);
  // Color
  toColorBuffer(0.25f, 1.0f, 0.1f);
  toColorBuffer(0.25f, 1.0f, 0.1f);
  toColorBuffer(0.25f, 1.0f, 0.1f);
  // 1.2
  // Normal
  vx = 0.05 - (-0.05);
  vy = 0.05 - 0.05;
  vz = 0.0 - 0.10;
  wx = 0.05 - (-0.05);
  wy = 0.05 - 0.05;
  wz = 0.10 - 0.10;
  computeNormal(&vx, &vy, &vz, wx, wy, wz);
  p = glm::vec4(vx, vy, vz, 1.0f);
  p = matrix * p;
  toNormalBuffer(p[0], p[1], p[2]);
  vx = -0.05 - 0.05;
  vy = 0.05 - 0.05;
  vz = 0.10 - 0.0;
  wx = 0.05 - 0.05;
  wy = 0.05 - 0.05;
  wz = 0.10 - 0.0;
  computeNormal(&vx, &vy, &vz, wx, wy, wz);
  p = glm::vec4(vx, vy, vz, 1.0f);
  p = matrix * p;
  toNormalBuffer(p[0], p[1], p[2]);
  vx = -0.05 - 0.05;
  vy = 0.05 - 0.05;
  vz = 0.10 - 0.10;
  wx = 0.05 - 0.05;
  wy = 0.05 - 0.05;
  wz = 0.0 - 0.10;
  computeNormal(&vx, &vy, &vz, wx, wy, wz);
  p = glm::vec4(vx, vy, vz, 1.0f);
  p = matrix * p;
  toNormalBuffer(p[0], p[1], p[2]);
  // Vertex
  p = glm::vec4(-0.05f, 0.05f, 1.0f, 1.0f);
  p = matrix * p;
  toVertexBuffer(p[0], p[1], p[2]);
  p = glm::vec4(0.05f, 0.05f, 0.0f, 1.0f);
  p = matrix * p;
  toVertexBuffer(p[0], p[1], p[2]);
  p = glm::vec4(0.05f, 0.05f, 0.10f, 1.0f);
  p = matrix * p;
  // Color
  toVertexBuffer(p[0], p[1], p[2]);
  toColorBuffer(0.25f, 1.0f, 0.1f);
  toColorBuffer(0.25f, 1.0f, 0.1f);
  toColorBuffer(0.25f, 1.0f, 0.1f);
  // 2.1
  // Vertex
  // Normal
  vx = 0.05 - 0.05;
  vy = -0.05 - 0.05;
  vz = 0.0 - 0.0;
  wx = 0.05 - 0.05;
  wy = 0.05 - 0.05;
  wz = 0.10 - 0.0;
  computeNormal(&vx, &vy, &vz, wx, wy, wz);
  p = glm::vec4(vx, vy, vz, 1.0f);
  p = matrix * p;
  toNormalBuffer(p[0], p[1], p[2]);
  vx = 0.05 - 0.05;
  vy = 0.05 - (-0.05);
  vz = 0.0 - 0.0;
  wx = 0.05 - 0.05;
  wy = 0.05 - (-0.05);
  wz = 0.10 - 0.0;
  computeNormal(&vx, &vy, &vz, wx, wy, wz);
  p = glm::vec4(vx, vy, vz, 1.0f);
  p = matrix * p;
  toNormalBuffer(p[0], p[1], p[2]);
  vx = 0.05 - 0.05;
  vy = 0.05 - 0.05;
  vz = 0.0 - 0.10;
  wx = 0.05 - 0.05;
  wy = -0.05 - 0.05;
  wz = 0.0 - 0.10;
  computeNormal(&vx, &vy, &vz, wx, wy, wz);
  p = glm::vec4(vx, vy, vz, 1.0f);
  p = matrix * p;
  toNormalBuffer(p[0], p[1], p[2]);
  p = glm::vec4(0.05f, 0.05f, 0.0f, 1.0f);
  p = matrix * p;
  toVertexBuffer(p[0], p[1], p[2]);
  p = glm::vec4(0.05f, -0.05f, 0.0f, 1.0f);
  p = matrix * p;
  toVertexBuffer(p[0], p[1], p[2]);
  p = glm::vec4(0.05f, 0.50f, 1.0f, 1.0f);
  p = matrix * p;
  toVertexBuffer(p[0], p[1], p[2]);
  // Color
  toColorBuffer(0.25f, 1.0f, 0.1f);
  toColorBuffer(0.25f, 1.0f, 0.1f);
  toColorBuffer(0.25f, 1.0f, 0.1f);
  // 2.2
  // Normal
  vx = 0.05 - 0.05;
  vy = 0.05 - (-0.05);
  vz = 0.10 - 0.0;
  wx = 0.05 - 0.05;
  wy = -0.05 - (-0.05);
  wz = 0.10 - 0.0;
  computeNormal(&vx, &vy, &vz, wx, wy, wz);
  p = glm::vec4(vx, vy, vz, 1.0f);
  p = matrix * p;
  toNormalBuffer(p[0], p[1], p[2]);
  vx = 0.05 - 0.05;
  vy = -0.05 - 0.05;
  vz = 0.0 - 0.10;
  wx = 0.05 - 0.05;
  wy = -0.05 - 0.05;
  wz = 0.10 - 0.10;
  computeNormal(&vx, &vy, &vz, wx, wy, wz);
  p = glm::vec4(vx, vy, vz, 1.0f);
  p = matrix * p;
  toNormalBuffer(p[0], p[1], p[2]);
  vx = 0.05 - 0.05;
  vy = -0.05 - (-0.05);
  vz = 0.0 - 0.10;
  wx = 0.05 - 0.05;
  wy = 0.05 - (-0.05);
  wz = 0.10 - 0.10;
  computeNormal(&vx, &vy, &vz, wx, wy, wz);
  p = glm::vec4(vx, vy, vz, 1.0f);
  p = matrix * p;
  toNormalBuffer(p[0], p[1], p[2]);
  // Vertex
  p = glm::vec4(0.05f, -0.05f, 0.0f, 1.0f);
  p = matrix * p;
  toVertexBuffer(p[0], p[1], p[2]);
  p = glm::vec4(0.05f, 0.05f, 1.0f, 1.0f);
  p = matrix * p;
  toVertexBuffer(p[0], p[1], p[2]);
  p = glm::vec4(0.05f, -0.05f, 1.0f, 1.0f);
  p = matrix * p;
  toVertexBuffer(p[0], p[1], p[2]);
  // Color
  toColorBuffer(0.25f, 1.0f, 0.1f);
  toColorBuffer(0.25f, 1.0f, 0.1f);
  toColorBuffer(0.25f, 1.0f, 0.1f);
  // 3.1
  // Normal
  vx = -0.05 - 0.05;
  vy = -0.05 - (-0.05);
  vz = 0.0 - 0.0;
  wx = 0.05 - 0.05;
  wy = -0.05 - (-0.05);
  wz = 0.10 - 0.0;
  computeNormal(&vx, &vy, &vz, wx, wy, wz);
  p = glm::vec4(vx, vy, vz, 1.0f);
  p = matrix * p;
  toNormalBuffer(p[0], p[1], p[2]);
  vx = 0.05 - (-0.05);
  vy = -0.05 - (-0.05);
  vz = 0.0 - 0.0;
  wx = 0.05 - (-0.05);
  wy = -0.05 - (-0.05);
  wz = 0.10 - 0.0;
  computeNormal(&vx, &vy, &vz, wx, wy, wz);
  p = glm::vec4(vx, vy, vz, 1.0f);
  p = matrix * p;
  toNormalBuffer(p[0], p[1], p[2]);
  vx = 0.05 - 0.05;
  vy = -0.05 - (-0.05);
  vz = 0.0 - 0.10;
  wx = -0.05 - 0.05;
  wy = -0.05 - (-0.05);
  wz = 0.0 - 0.10;
  computeNormal(&vx, &vy, &vz, wx, wy, wz);
  p = glm::vec4(vx, vy, vz, 1.0f);
  p = matrix * p;
  toNormalBuffer(p[0], p[1], p[2]);
  // Vertex
  p = glm::vec4(0.05f, -0.05f, 0.0f, 1.0f);
  p = matrix * p;
  toVertexBuffer(p[0], p[1], p[2]);
  p = glm::vec4(-0.05f, -0.05f, 0.0f, 1.0f);
  p = matrix * p;
  toVertexBuffer(p[0], p[1], p[2]);
  p = glm::vec4(0.05f, -0.05f, 1.0f, 1.0f);
  p = matrix * p;
  toVertexBuffer(p[0], p[1], p[2]);
  // Color
  toColorBuffer(0.25f, 1.0f, 0.1f);
  toColorBuffer(0.25f, 1.0f, 0.1f);
  toColorBuffer(0.25f, 1.0f, 0.1f);
  // 3.2
  // Normal
  vx = 0.5 - (-0.5);
  vy = -0.5 - (-0.5);
  vz = 1.0 - 0.0;
  wx = -0.5 - (-0.5);
  wy = -0.5 - (-0.5);
  wz = 1.0 - 0.0;
  computeNormal(&vx, &vy, &vz, wx, wy, wz);
  p = glm::vec4(vx, vy, vz, 1.0f);
  p = matrix * p;
  toNormalBuffer(p[0], p[1], p[2]);
  vx = -0.5 - 0.5;
  vy = -0.5 - (-0.5);
  vz = 0.0 - 1.0;
  wx = -0.5 - 0.5;
  wy = -0.5 - (-0.5);
  wz = 1.0 - 1.0;
  computeNormal(&vx, &vy, &vz, wx, wy, wz);
  p = glm::vec4(vx, vy, vz, 1.0f);
  p = matrix * p;
  toNormalBuffer(p[0], p[1], p[2]);
  vx = -0.5 - (-0.5);
  vy = -0.5 - (-0.5);
  vz = 0.0 - 1.0;
  wx = 0.5 - (-0.5);
  wy = -0.5 - (-0.5);
  wz = 1.0 - 1.0;
  computeNormal(&vx, &vy, &vz, wx, wy, wz);
  p = glm::vec4(vx, vy, vz, 1.0f);
  p = matrix * p;
  toNormalBuffer(p[0], p[1], p[2]);
  // Vertex
  p = glm::vec4(-0.05f, -0.05f, 0.0f, 1.0f);
  p = matrix * p;
  toVertexBuffer(p[0], p[1], p[2]);
  p = glm::vec4(0.05f, -0.05f, 1.0f, 1.0f);
  p = matrix * p;
  toVertexBuffer(p[0], p[1], p[2]);
  p = glm::vec4(-0.05f, -0.05f, 1.0f, 1.0f);
  p = matrix * p;
  toVertexBuffer(p[0], p[1], p[2]);
  // Color
  toColorBuffer(0.25f, 1.0f, 0.1f);
  toColorBuffer(0.25f, 1.0f, 0.1f);
  toColorBuffer(0.25f, 1.0f, 0.1f);
  // 4.1
  // Normal
  vx = -0.5 - (-0.5);
  vy = 0.5 - (-0.5);
  vz = 0.0 - 0.0;
  wx = -0.5 - (-0.5);
  wy = -0.5 - (-0.5);
  wz = 1.0 - 0.0;
  computeNormal(&vx, &vy, &vz, wx, wy, wz);
  p = glm::vec4(vx, vy, vz, 1.0f);
  p = matrix * p;
  toNormalBuffer(p[0], p[1], p[2]);
  vx = -0.5 - (-0.5);
  vy = -0.5 - 0.5;
  vz = 0.0 - 0.0;
  wx = -0.5 - (-0.5);
  wy = -0.5 - 0.5;
  wz = 1.0 - 0.0;
  computeNormal(&vx, &vy, &vz, wx, wy, wz);
  p = glm::vec4(vx, vy, vz, 1.0f);
  p = matrix * p;
  toNormalBuffer(p[0], p[1], p[2]);
  vx = -0.5 - (-0.5);
  vy = -0.5 - (-0.5);
  vz = 0.0 - 1.0;
  wx = -0.5 - (-0.5);
  wy = 0.5 - (-0.5);
  wz = 0.0 - 1.0;
  computeNormal(&vx, &vy, &vz, wx, wy, wz);
  p = glm::vec4(vx, vy, vz, 1.0f);
  p = matrix * p;
  toNormalBuffer(p[0], p[1], p[2]);
  // Vertex
  p = glm::vec4(-0.05f, -0.05f, 0.0f, 1.0f);
  p = matrix * p;
  toVertexBuffer(p[0], p[1], p[2]);
  p = glm::vec4(-0.05f, 0.05f, 0.0f, 1.0f);
  p = matrix * p;
  toVertexBuffer(p[0], p[1], p[2]);
  p = glm::vec4(-0.05f, -0.05f, 1.0f, 1.0f);
  p = matrix * p;
  toVertexBuffer(p[0], p[1], p[2]);
  // Color
  toColorBuffer(0.25f, 1.0f, 0.1f);
  toColorBuffer(0.25f, 1.0f, 0.1f);
  toColorBuffer(0.25f, 1.0f, 0.1f);
  // 4.2
  // Normal
  vx = -0.5 - (-0.5);
  vy = -0.5 - 0.5;
  vz = 1.0 - 0.0;
  wx = -0.5 - (-0.5);
  wy = 0.5 - 0.5;
  wz = 1.0 - 0.0;
  computeNormal(&vx, &vy, &vz, wx, wy, wz);
  p = glm::vec4(vx, vy, vz, 1.0f);
  p = matrix * p;
  toNormalBuffer(p[0], p[1], p[2]);
  vx = -0.5 - (-0.5);
  vy = 0.5 - (-0.5);
  vz = 0.0 - 1.0;
  wx = -0.5 - (-0.5);
  wy = 0.5 - (-0.5);
  wz = 1.0 - 1.0;
  computeNormal(&vx, &vy, &vz, wx, wy, wz);
  p = glm::vec4(vx, vy, vz, 1.0f);
  p = matrix * p;
  toNormalBuffer(p[0], p[1], p[2]);
  vx = -0.5 - (-0.5);
  vy = 0.5 - 0.5;
  vz = 0.0 - 1.0;
  wx = -0.5 - (-0.5);
  wy = -0.5 - 0.5;
  wz = 1.0 - 1.0;
  computeNormal(&vx, &vy, &vz, wx, wy, wz);
  p = glm::vec4(vx, vy, vz, 1.0f);
  p = matrix * p;
  toNormalBuffer(p[0], p[1], p[2]);
  // Vertex
  p = glm::vec4(-0.05f, 0.05f, 0.0f, 1.0f);
  p = matrix * p;
  toVertexBuffer(p[0], p[1], p[2]);
  p = glm::vec4(-0.05f, -0.05f, 1.0f, 1.0f);
  p = matrix * p;
  toVertexBuffer(p[0], p[1], p[2]);
  p = glm::vec4(-0.05f, 0.05f, 1.0f, 1.0f);
  p = matrix * p;
  toVertexBuffer(p[0], p[1], p[2]);
  // Color
  toColorBuffer(0.25f, 1.0f, 0.1f);
  toColorBuffer(0.25f, 1.0f, 0.1f);
  toColorBuffer(0.25f, 1.0f, 0.1f);

  numTri += 8;
}

// Draws a single leaf, along the current local Z axis
void leafSection(glm::mat4 matrix) {
  glm::vec4 p;

  stemSection(matrix);
  // Leaf
  // 1
  matrix = matrix * glm::scale(glm::mat4(1.0), glm::vec3(0.5));
  p = glm::vec4(0.0f, 0.0f, 1.0f, 1.0f);
  p = matrix * p;
  toNormalBuffer(p[0], p[1], p[2]);
  p = glm::vec4(0.0f, 0.0f, 0.0f, 1.0f);
  p = matrix * p;
  toVertexBuffer(p[0], p[1], p[2]);
  p = glm::vec4(0.0f, 0.0f, 1.0f, 1.0f);
  p = matrix * p;
  toNormalBuffer(p[0], p[1], p[2]);
  p = glm::vec4(-0.06f, -0.2f, 0.0f, 1.0f);
  p = matrix * p;
  toVertexBuffer(p[0], p[1], p[2]);
  p = glm::vec4(0.0f, 0.0f, 1.0f, 1.0f);
  p = matrix * p;
  toNormalBuffer(p[0], p[1], p[2]);
  p = glm::vec4(0.06f, -0.2f, 0.0f, 1.0f);
  p = matrix * p;
  toVertexBuffer(p[0], p[1], p[2]);
  toColorBuffer(0.25f, 1.0f, 0.1f);
  toColorBuffer(0.25f, 1.0f, 0.1f);
  toColorBuffer(0.25f, 1.0f, 0.1f);
  // 2
  p = glm::vec4(0.0f, 0.0f, 1.0f, 1.0f);
  p = matrix * p;
  toNormalBuffer(p[0], p[1], p[2]);
  p = glm::vec4(-0.06f, -0.2f, 0.0f, 1.0f);
  p = matrix * p;
  toVertexBuffer(p[0], p[1], p[2]);
  p = glm::vec4(0.0f, 0.0f, 1.0f, 1.0f);
  p = matrix * p;
  toNormalBuffer(p[0], p[1], p[2]);
  p = glm::vec4(-0.07f, -0.3f, 0.0f, 1.0f);
  p = matrix * p;
  toVertexBuffer(p[0], p[1], p[2]);
  p = glm::vec4(0.0f, 0.0f, 1.0f, 1.0f);
  p = matrix * p;
  toNormalBuffer(p[0], p[1], p[2]);
  p = glm::vec4(0.07f, -0.3f, 0.0f, 1.0f);
  p = matrix * p;
  toVertexBuffer(p[0], p[1], p[2]);
  toColorBuffer(0.25f, 1.0f, 0.1f);
  toColorBuffer(0.25f, 1.0f, 0.1f);
  toColorBuffer(0.25f, 1.0f, 0.1f);
  // 3
  p = glm::vec4(0.0f, 0.0f, 1.0f, 1.0f);
  p = matrix * p;
  toNormalBuffer(p[0], p[1], p[2]);
  p = glm::vec4(-0.06f, -0.2f, 0.0f, 1.0f);
  p = matrix * p;
  toVertexBuffer(p[0], p[1], p[2]);
  p = glm::vec4(0.0f, 0.0f, 1.0f, 1.0f);
  p = matrix * p;
  toNormalBuffer(p[0], p[1], p[2]);
  p = glm::vec4(0.06f, -0.2f, 0.0f, 1.0f);
  p = matrix * p;
  toVertexBuffer(p[0], p[1], p[2]);
  p = glm::vec4(0.0f, 0.0f, 1.0f, 1.0f);
  p = matrix * p;
  toNormalBuffer(p[0], p[1], p[2]);
  p = glm::vec4(0.07f, -0.3f, 0.0f, 1.0f);
  p = matrix * p;
  toVertexBuffer(p[0], p[1], p[2]);
  toColorBuffer(0.25f, 1.0f, 0.1f);
  toColorBuffer(0.25f, 1.0f, 0.1f);
  toColorBuffer(0.25f, 1.0f, 0.1f);
  // 4
  p = glm::vec4(0.0f, 0.0f, 1.0f, 1.0f);
  p = matrix * p;
  toNormalBuffer(p[0], p[1], p[2]);
  p = glm::vec4(-0.07f, -0.3f, 0.0f, 1.0f);
  p = matrix * p;
  toVertexBuffer(p[0], p[1], p[2]);
  p = glm::vec4(0.0f, 0.0f, 1.0f, 1.0f);
  p = matrix * p;
  toNormalBuffer(p[0], p[1], p[2]);
  p = glm::vec4(-0.06f, -0.4f, 0.0f, 1.0f);
  p = matrix * p;
  toVertexBuffer(p[0], p[1], p[2]);
  p = glm::vec4(0.0f, 0.0f, 1.0f, 1.0f);
  p = matrix * p;
  toNormalBuffer(p[0], p[1], p[2]);
  p = glm::vec4(0.06f, -0.4f, 0.0f, 1.0f);
  p = matrix * p;
  toVertexBuffer(p[0], p[1], p[2]);
  toColorBuffer(0.25f, 1.0f, 0.1f);
  toColorBuffer(0.25f, 1.0f, 0.1f);
  toColorBuffer(0.25f, 1.0f, 0.1f);
  // 5
  p = glm::vec4(0.0f, 0.0f, 1.0f, 1.0f);
  p = matrix * p;
  toNormalBuffer(p[0], p[1], p[2]);
  p = glm::vec4(-0.07f, -0.3f, 0.0f, 1.0f);
  p = matrix * p;
  toVertexBuffer(p[0], p[1], p[2]);
  p = glm::vec4(0.0f, 0.0f, 1.0f, 1.0f);
  p = matrix * p;
  toNormalBuffer(p[0], p[1], p[2]);
  p = glm::vec4(0.07f, -0.3f, 0.0f, 1.0f);
  p = matrix * p;
  toVertexBuffer(p[0], p[1], p[2]);
  p = glm::vec4(0.0f, 0.0f, 1.0f, 1.0f);
  p = matrix * p;
  toNormalBuffer(p[0], p[1], p[2]);
  p = glm::vec4(0.06f, -0.4f, 0.0f, 1.0f);
  p = matrix * p;
  toVertexBuffer(p[0], p[1], p[2]);
  toColorBuffer(0.25f, 1.0f, 0.1f);
  toColorBuffer(0.25f, 1.0f, 0.1f);
  toColorBuffer(0.25f, 1.0f, 0.1f);
  // 6
  p = glm::vec4(0.0f, 0.0f, 1.0f, 1.0f);
  p = matrix * p;
  toNormalBuffer(p[0], p[1], p[2]);
  p = glm::vec4(-0.06f, -0.4f, 0.0f, 1.0f);
  p = matrix * p;
  toVertexBuffer(p[0], p[1], p[2]);
  p = glm::vec4(0.0f, 0.0f, 1.0f, 1.0f);
  p = matrix * p;
  toNormalBuffer(p[0], p[1], p[2]);
  p = glm::vec4(-0.04f, -0.5f, 0.0f, 1.0f);
  p = matrix * p;
  toVertexBuffer(p[0], p[1], p[2]);
  p = glm::vec4(0.0f, 0.0f, 1.0f, 1.0f);
  p = matrix * p;
  toNormalBuffer(p[0], p[1], p[2]);
  p = glm::vec4(0.04f, -0.5f, 0.0f, 1.0f);
  p = matrix * p;
  toVertexBuffer(p[0], p[1], p[2]);
  toColorBuffer(0.25f, 1.0f, 0.1f);
  toColorBuffer(0.25f, 1.0f, 0.1f);
  toColorBuffer(0.25f, 1.0f, 0.1f);
  // 7
  p = glm::vec4(0.0f, 0.0f, 1.0f, 1.0f);
  p = matrix * p;
  toNormalBuffer(p[0], p[1], p[2]);
  p = glm::vec4(-0.06f, -0.4f, 0.0f, 1.0f);
  p = matrix * p;
  toVertexBuffer(p[0], p[1], p[2]);
  p = glm::vec4(0.0f, 0.0f, 1.0f, 1.0f);
  p = matrix * p;
  toNormalBuffer(p[0], p[1], p[2]);
  p = glm::vec4(0.06f, -0.4f, 0.0f, 1.0f);
  p = matrix * p;
  toVertexBuffer(p[0], p[1], p[2]);
  p = glm::vec4(0.0f, 0.0f, 1.0f, 1.0f);
  p = matrix * p;
  toNormalBuffer(p[0], p[1], p[2]);
  p = glm::vec4(0.04f, -0.5f, 0.0f, 1.0f);
  p = matrix * p;
  toVertexBuffer(p[0], p[1], p[2]);
  toColorBuffer(0.25f, 1.0f, 0.1f);
  toColorBuffer(0.25f, 1.0f, 0.1f);
  toColorBuffer(0.25f, 1.0f, 0.1f);
  // 8
  p = glm::vec4(0.0f, 0.0f, 1.0f, 1.0f);
  p = matrix * p;
  toNormalBuffer(p[0], p[1], p[2]);
  p = glm::vec4(-0.04f, -0.5f, 0.0f, 1.0f);
  p = matrix * p;
  toVertexBuffer(p[0], p[1], p[2]);
  p = glm::vec4(0.0f, 0.0f, 1.0f, 1.0f);
  p = matrix * p;
  toNormalBuffer(p[0], p[1], p[2]);
  p = glm::vec4(0.04f, -0.5f, 0.0f, 1.0f);
  p = matrix * p;
  toVertexBuffer(p[0], p[1], p[2]);
  p = glm::vec4(0.0f, 0.0f, 1.0f, 1.0f);
  p = matrix * p;
  toNormalBuffer(p[0], p[1], p[2]);
  p = glm::vec4(0.0f, -0.53f, 0.0f, 1.0f);
  p = matrix * p;
  toVertexBuffer(p[0], p[1], p[2]);
  toColorBuffer(0.25f, 1.0f, 0.1f);
  toColorBuffer(0.25f, 1.0f, 0.1f);
  toColorBuffer(0.25f, 1.0f, 0.1f);

  numTri += 8;
}

// Draws a flower perpendicular to the current local Z axis
void flowerSection(glm::mat4 matrix) {
  glm::vec4 p;
  matrix = matrix * glm::scale(glm::mat4(1.0), glm::vec3(0.7));
  // Rotates 8 leafs to get a flower
  for (int i = 0; i < 6; i++) {
    // Petals
    matrix = matrix * glm::rotate(glm::mat4(1.0), 45.0f, glm::vec3(0.0f, 0.0f, 1.0f));
    // 1
    p = glm::vec4(0.0f, 0.0f, 1.0f, 1.0f);
    p = matrix * p;
    toNormalBuffer(p[0], p[1], p[2]);
    p = glm::vec4(0.0f, 0.0f, 0.0f, 1.0f);
    p = matrix * p;
    toVertexBuffer(p[0], p[1], p[2]);
    p = glm::vec4(0.0f, 0.0f, 1.0f, 1.0f);
    p = matrix * p;
    toNormalBuffer(p[0], p[1], p[2]);
    p = glm::vec4(-0.06f, -0.2f, 0.0f, 1.0f);
    p = matrix * p;
    toVertexBuffer(p[0], p[1], p[2]);
    p = glm::vec4(0.0f, 0.0f, 1.0f, 1.0f);
    p = matrix * p;
    toNormalBuffer(p[0], p[1], p[2]);
    p = glm::vec4(0.06f, -0.2f, 0.0f, 1.0f);
    p = matrix * p;
    toVertexBuffer(p[0], p[1], p[2]);
    toColorBuffer(1.0f, 0.0f, 0.0f);
    toColorBuffer(1.0f, 0.0f, 0.0f);
    toColorBuffer(1.0f, 0.0f, 0.0f);
    // 2
    p = glm::vec4(0.0f, 0.0f, 1.0f, 1.0f);
    p = matrix * p;
    toNormalBuffer(p[0], p[1], p[2]);
    p = glm::vec4(-0.06f, -0.2f, 0.0f, 1.0f);
    p = matrix * p;
    toVertexBuffer(p[0], p[1], p[2]);
    p = glm::vec4(0.0f, 0.0f, 1.0f, 1.0f);
    p = matrix * p;
    toNormalBuffer(p[0], p[1], p[2]);
    p = glm::vec4(-0.07f, -0.3f, 0.0f, 1.0f);
    p = matrix * p;
    toVertexBuffer(p[0], p[1], p[2]);
    p = glm::vec4(0.0f, 0.0f, 1.0f, 1.0f);
    p = matrix * p;
    toNormalBuffer(p[0], p[1], p[2]);
    p = glm::vec4(0.07f, -0.3f, 0.0f, 1.0f);
    p = matrix * p;
    toVertexBuffer(p[0], p[1], p[2]);
    toColorBuffer(1.0f, 0.0f, 0.0f);
    toColorBuffer(1.0f, 0.0f, 0.0f);
    toColorBuffer(1.0f, 0.0f, 0.0f);
    // 3
    p = glm::vec4(0.0f, 0.0f, 1.0f, 1.0f);
    p = matrix * p;
    toNormalBuffer(p[0], p[1], p[2]);
    p = glm::vec4(-0.06f, -0.2f, 0.0f, 1.0f);
    p = matrix * p;
    toVertexBuffer(p[0], p[1], p[2]);
    p = glm::vec4(0.0f, 0.0f, 1.0f, 1.0f);
    p = matrix * p;
    toNormalBuffer(p[0], p[1], p[2]);
    p = glm::vec4(0.06f, -0.2f, 0.0f, 1.0f);
    p = matrix * p;
    toVertexBuffer(p[0], p[1], p[2]);
    p = glm::vec4(0.0f, 0.0f, 1.0f, 1.0f);
    p = matrix * p;
    toNormalBuffer(p[0], p[1], p[2]);
    p = glm::vec4(0.07f, -0.3f, 0.0f, 1.0f);
    p = matrix * p;
    toVertexBuffer(p[0], p[1], p[2]);
    toColorBuffer(1.0f, 0.0f, 0.0f);
    toColorBuffer(1.0f, 0.0f, 0.0f);
    toColorBuffer(1.0f, 0.0f, 0.0f);
    // 4
    p = glm::vec4(0.0f, 0.0f, 1.0f, 1.0f);
    p = matrix * p;
    toNormalBuffer(p[0], p[1], p[2]);
    p = glm::vec4(-0.07f, -0.3f, 0.0f, 1.0f);
    p = matrix * p;
    toVertexBuffer(p[0], p[1], p[2]);
    p = glm::vec4(0.0f, 0.0f, 1.0f, 1.0f);
    p = matrix * p;
    toNormalBuffer(p[0], p[1], p[2]);
    p = glm::vec4(-0.06f, -0.4f, 0.0f, 1.0f);
    p = matrix * p;
    toVertexBuffer(p[0], p[1], p[2]);
    p = glm::vec4(0.0f, 0.0f, 1.0f, 1.0f);
    p = matrix * p;
    toNormalBuffer(p[0], p[1], p[2]);
    p = glm::vec4(0.06f, -0.4f, 0.0f, 1.0f);
    p = matrix * p;
    toVertexBuffer(p[0], p[1], p[2]);
    toColorBuffer(1.0f, 0.0f, 0.0f);
    toColorBuffer(1.0f, 0.0f, 0.0f);
    toColorBuffer(1.0f, 0.0f, 0.0f);
    // 5
    p = glm::vec4(0.0f, 0.0f, 1.0f, 1.0f);
    p = matrix * p;
    toNormalBuffer(p[0], p[1], p[2]);
    p = glm::vec4(-0.07f, -0.3f, 0.0f, 1.0f);
    p = matrix * p;
    toVertexBuffer(p[0], p[1], p[2]);
    p = glm::vec4(0.0f, 0.0f, 1.0f, 1.0f);
    p = matrix * p;
    toNormalBuffer(p[0], p[1], p[2]);
    p = glm::vec4(0.07f, -0.3f, 0.0f, 1.0f);
    p = matrix * p;
    toVertexBuffer(p[0], p[1], p[2]);
    p = glm::vec4(0.0f, 0.0f, 1.0f, 1.0f);
    p = matrix * p;
    toNormalBuffer(p[0], p[1], p[2]);
    p = glm::vec4(0.06f, -0.4f, 0.0f, 1.0f);
    p = matrix * p;
    toVertexBuffer(p[0], p[1], p[2]);
    toColorBuffer(1.0f, 0.0f, 0.0f);
    toColorBuffer(1.0f, 0.0f, 0.0f);
    toColorBuffer(1.0f, 0.0f, 0.0f);
    // 6
    p = glm::vec4(0.0f, 0.0f, 1.0f, 1.0f);
    p = matrix * p;
    toNormalBuffer(p[0], p[1], p[2]);
    p = glm::vec4(-0.06f, -0.4f, 0.0f, 1.0f);
    p = matrix * p;
    toVertexBuffer(p[0], p[1], p[2]);
    p = glm::vec4(0.0f, 0.0f, 1.0f, 1.0f);
    p = matrix * p;
    toNormalBuffer(p[0], p[1], p[2]);
    p = glm::vec4(-0.04f, -0.5f, 0.0f, 1.0f);
    p = matrix * p;
    toVertexBuffer(p[0], p[1], p[2]);
    p = glm::vec4(0.0f, 0.0f, 1.0f, 1.0f);
    p = matrix * p;
    toNormalBuffer(p[0], p[1], p[2]);
    p = glm::vec4(0.04f, -0.5f, 0.0f, 1.0f);
    p = matrix * p;
    toVertexBuffer(p[0], p[1], p[2]);
    toColorBuffer(1.0f, 0.0f, 0.0f);
    toColorBuffer(1.0f, 0.0f, 0.0f);
    toColorBuffer(1.0f, 0.0f, 0.0f);
    // 7
    p = glm::vec4(0.0f, 0.0f, 1.0f, 1.0f);
    p = matrix * p;
    toNormalBuffer(p[0], p[1], p[2]);
    p = glm::vec4(-0.06f, -0.4f, 0.0f, 1.0f);
    p = matrix * p;
    toVertexBuffer(p[0], p[1], p[2]);
    p = glm::vec4(0.0f, 0.0f, 1.0f, 1.0f);
    p = matrix * p;
    toNormalBuffer(p[0], p[1], p[2]);
    p = glm::vec4(0.06f, -0.4f, 0.0f, 1.0f);
    p = matrix * p;
    toVertexBuffer(p[0], p[1], p[2]);
    p = glm::vec4(0.0f, 0.0f, 1.0f, 1.0f);
    p = matrix * p;
    toNormalBuffer(p[0], p[1], p[2]);
    p = glm::vec4(0.04f, -0.5f, 0.0f, 1.0f);
    p = matrix * p;
    toVertexBuffer(p[0], p[1], p[2]);
    toColorBuffer(1.0f, 0.0f, 0.0f);
    toColorBuffer(1.0f, 0.0f, 0.0f);
    toColorBuffer(1.0f, 0.0f, 0.0f);
    // 8
    p = glm::vec4(0.0f, 0.0f, 1.0f, 1.0f);
    p = matrix * p;
    toNormalBuffer(p[0], p[1], p[2]);
    p = glm::vec4(-0.04f, -0.5f, 0.0f, 1.0f);
    p = matrix * p;
    toVertexBuffer(p[0], p[1], p[2]);
    p = glm::vec4(0.0f, 0.0f, 1.0f, 1.0f);
    p = matrix * p;
    toNormalBuffer(p[0], p[1], p[2]);
    p = glm::vec4(0.04f, -0.5f, 0.0f, 1.0f);
    p = matrix * p;
    toVertexBuffer(p[0], p[1], p[2]);
    p = glm::vec4(0.0f, 0.0f, 1.0f, 1.0f);
    p = matrix * p;
    toNormalBuffer(p[0], p[1], p[2]);
    p = glm::vec4(0.0f, -0.53f, 0.0f, 1.0f);
    p = matrix * p;
    toVertexBuffer(p[0], p[1], p[2]);
    toColorBuffer(1.0f, 0.0f, 0.0f);
    toColorBuffer(1.0f, 0.0f, 0.0f);
    toColorBuffer(1.0f, 0.0f, 0.0f);
  }

  numTri += 8 * 8;
}

// Release all memory allocated to the plant structure
void freePlant(struct PlantNode *p) {
  if (p == NULL) 
    return;
  if (p->left != NULL) 
    freePlant(p->left);
  if (p->right != NULL) 
    freePlant(p->right);
  free(p);
}

// Tree traversal to print the plant structure
// To enable viewing the structure of the plant,
// Nodes are printed by level
void printPlant(struct PlantNode *p) {
  int l;
  for (l = 0; l < n_levels; l++) {
    fprintf(stderr, "Level %02d\n", l);
    printNodeRecursive(p, 0, l);
    fprintf(stderr, "\n");
  }
}

void printNodeRecursive(struct PlantNode *p, int lev, int tgt) {
  if (p == NULL) 
    return;
  if (lev == tgt) {
    fprintf(stderr, "%c ", p->type);
    return;
  }
  if (p->left != NULL) 
    printNodeRecursive(p->left, lev + 1, tgt);
  if (p->right != NULL) 
    printNodeRecursive(p->right, lev + 1, tgt);
  fprintf(stderr, " ");
  return;
}

// This function creates a plant with the specified
// characteristics given by the global plant parameters
// and a set of internal parameters that govern the
// probability of specific parts being generated
// at each level.
//
// The rules for generating the plant are as follows:
//
// Four types of components:  a - Plant stem
//                            b - Plant stem
//                            c - Leaf
//                            d - flower
//
// The generation rules for the plant structure are
//
// a -> aa with p = Paaa
// a -> ab with p = Paab
// a -> ac with p = Paac
// a -> ad with p = Paad
// a -> cd with p = Pacd
// b -> a  with p = Pba
// b -> c  with p = Pbc
// b -> d  with p = Pbd
// c and d are terminal nodes.
//
// The probabilities for transitions are parameters to
// the function so that you can generate different
// types of plants.
struct PlantNode *makePlant(void) {
  struct PlantNode *p_root = NULL;

  p_root = (struct PlantNode *) calloc(1, sizeof(struct PlantNode)); // Allocate new tree root
  p_root->type = 'a';                                              // Always start with a stem
  p_root->z_ang = 0;                                               // which is vertical w.r.t
  p_root->x_ang = 0;                                               // global coordinate frame
  p_root->left = NULL;
  p_root->left = NULL;

  // Generative part, use the laws above to generate up to two children for each node in the tree
  generateRecursivePlant(p_root, 1);  // Level 0 is the root, next level is 1
  return(p_root);
}

// Generates child nodes for parent 'p' with p being at level 'level'
// if level > n_levels, stop.
void generateRecursivePlant(struct PlantNode *p, int level) {
  float dice;
  struct PlantNode *q, *r;

  q = r = NULL;
  if (p == NULL) return;                     // Reached a terminal
  if (level >= n_levels) return;             // Reached maximum plant height
  if (p->type == 'c' || p->type == 'd') return;  // c and d type nodes are terminal nodes as well

  dice = rand48();           // Roll the dice...
  if (p->type == 'a') {
    // Generate nodes for left and right
    q = (struct PlantNode *)calloc(1, sizeof(struct PlantNode));
    q->x_ang = rand48() * X_angle;
    q->z_ang = rand48() * Z_angle;
    q->left = NULL;
    q->right = NULL;
    r = (struct PlantNode *) calloc(1, sizeof(struct PlantNode));
    r->x_ang = rand48() * X_angle;
    r->z_ang = rand48() * Z_angle;
    r->left = NULL;
    r->right = NULL;

    if (dice <= Paab) {
      // Selected rule a -> ab
      q->type = 'a';
      r->type = 'b';
    } else if (dice <= (Paab + Paac)) {
      // Selected rule a -> ac
      q->type = 'a';
      r->type = 'c';
    } else if (dice <= (Paab + Paac + Paad)) {
      // Selected rule a -> ad
      q->type = 'a';
      r->type = 'd';
    } else if (dice <= (Paab + Paac + Paad + Pacd)) {
      // Selected rule a -> cd
      q->type = 'c';
      r->type = 'd';
    } else {
      // Selected rule a -> aa
      q->type = 'a';
      r->type = 'a';
    }
  } else if (p->type == 'b') {
    // Generate a single node for either left or right
    q = NULL;
    r = (struct PlantNode *) calloc(1, sizeof(struct PlantNode));
    r->x_ang = rand48() * X_angle;
    r->z_ang = rand48() * Z_angle;
    r->left = NULL;
    r->right = NULL;

    if (dice <= Pba) {
      // Selected rule b -> a
      r->type = 'a';
    } else if (dice <= (Pba + Pbc)) {
      // Selected rule b -> c
      r->type = 'c';
    } else {
      // Selected rule b -> d
      r->type = 'd';
    }
  } else {
    fprintf(stderr, "Bad node type!\n");
    return;
  }

  // Decide which node goes to left and which goes right randomly
  if (rand48() <= .5) {
    p->left = q;
    p->right = r;
  } else {
    p->left = r;
    p->right = q;
  }

  // Recursive call for children
  generateRecursivePlant(p->left, level + 1);
  generateRecursivePlant(p->right, level + 1);
}

void Initialize()
{
  indexNormal = indexVertex = indexColor = 0;
  numTri = 0;
  /* Plants */
  n_levels = 3;
  X_angle = 30;
  Z_angle = 30;
  Paab = 0.15;
  Paac = 0.15;
  Paad = 0.15;
  Pacd = 0.15;
  Pba = 0.15;
  Pbc = 0.15;
  Pbd = 0.10;
  // Generate surface map
  makeSurfaceGrid();
  toBuffers();
  // Make a plant forest!
  for (int i = 0; i < MAX_PLANTS; i++) {
    struct PlantNode *plant = makePlant();
    int x = rand() % GRID_RESOLVE;
    int y = rand() % GRID_RESOLVE;
    glm::mat4 matrix = glm::translate(glm::mat4(), glm::vec3(GroundXYZ[x][y][0], GroundXYZ[x][y][1], GroundXYZ[x][y][2] - 10));
    toBufferPlant(plant, matrix);
    freePlant(plant);
  }

  tc_matrix = glm::translate(glm::mat4(1.0f), glm::vec3(0.0f, 0.0f, -3.0f)) * glm::scale(glm::mat4(1.0f), glm::vec3(0.3f));
  
  // Create the program for rendering the scene from the light's POV. 
  glCreateProgram();

  ShaderInfo shaders_info[] =
    {
      { GL_VERTEX_SHADER, "shaders/projeto.vs.glsl" },
      { GL_FRAGMENT_SHADER, "shaders/projeto.fs.glsl" },
      { GL_NONE, NULL }
    };

  shaders = LoadShaders(shaders_info);

  mvMatrix = glGetUniformLocation(shaders, "MVMatrix");
  prjMatrix = glGetUniformLocation(shaders, "PMatrix");
		
  // Create a vertex array object
  glGenVertexArrays(1, &vao);
  glBindVertexArray(vao);
  
  glGenBuffers(1, &vertexbuffer);
  glBindBuffer(GL_ARRAY_BUFFER, vertexbuffer);
  glBufferData(GL_ARRAY_BUFFER, sizeof(vertexbuffer_data), vertexbuffer_data, GL_STATIC_DRAW);
  // Tell OpenGL the format of your array data
  glVertexAttribPointer(
			0,                                // attribute. No particular reason for 0, but must match the layout in the shader.
			3,                                // size
			GL_FLOAT,                         // type
			GL_FALSE,                         // normalized?
			0,                                // stride
			(void*)0                          // array buffer offset
			);

  glGenBuffers(1, &colorbuffer);
  glBindBuffer(GL_ARRAY_BUFFER, colorbuffer);
  glBufferData(GL_ARRAY_BUFFER, sizeof(colorbuffer_data), colorbuffer_data, GL_STATIC_DRAW);
  // Tell OpenGL the format of your array data
  glVertexAttribPointer(
			1,                                // attribute. No particular reason for 1, but must match the layout in the shader.
			3,                                // size
			GL_FLOAT,                         // type
			GL_FALSE,                         // normalized?
			0,                                // stride
			(void*)0                          // array buffer offset
			);

  glGenBuffers(1, &normalbuffer);
  glBindBuffer(GL_ARRAY_BUFFER, normalbuffer);
  glBufferData(GL_ARRAY_BUFFER, sizeof(normalbuffer_data), normalbuffer_data, GL_STATIC_DRAW);
  // Tell OpenGL the format of your array data
  glVertexAttribPointer(
			2,                                // attribute. No particular reason for 2, but must match the layout in the shader.
			3,                                // size
			GL_FLOAT,                         // type
			GL_FALSE,                         // normalized?
			0,                                // stride
			(void*)0                          // array buffer offset
			);

  glClearColor(0.0f, 0.0f, 0.0f, 1.0f);
  glClearDepth(1.0f);
  glEnable(GL_DEPTH_TEST);
  glDepthFunc(GL_LEQUAL);
  glDisable(GL_CULL_FACE);
}

void Display(void)
{
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

  // Activate shader program
  glUseProgram(shaders);
  glClear(GL_DEPTH_BUFFER_BIT);

  glUniformMatrix4fv(prjMatrix, 1, GL_FALSE, glm::value_ptr(glm::perspective(30.0f, 1.0f / aspect, 0.1f, 10.0f)));
  

  /* ArcBall Code */
  if (curX != lastX || curY != lastY) {
    glm::vec3 va = get_arcball_vector(lastX, lastY);
    glm::vec3 vb = get_arcball_vector(curX, curY);
    float angle = acos(min(1.0f, glm::dot(va, vb)));
    glm::vec3 axis_in_camera_coord = glm::cross(va, vb);
    glm::mat3 camera2object = glm::inverse(glm::mat3(glm::perspective(30.0f, 1.0f / aspect, 0.1f, 10.0f))) * (glm::mat3(tc_matrix));
    glm::vec3 axis_in_object_coord = camera2object * axis_in_camera_coord;
    tc_matrix = glm::rotate(tc_matrix, glm::degrees(angle), axis_in_object_coord);
    lastX = curX;
    lastY = curY;
  }
  glUniformMatrix4fv(mvMatrix, 1, GL_FALSE, glm::value_ptr(tc_matrix));
  
  // Enable the vertex attributes for rendering
  glBindVertexArray(vao);
  glEnableVertexAttribArray(0);
  glEnableVertexAttribArray(1);
  glEnableVertexAttribArray(2);
  glDrawArrays(GL_TRIANGLES, 0, numTri);

  glutSwapBuffers();
  glutPostRedisplay();
}

void Keyboard(unsigned char key, int x, int y)
{
  switch (key) {
  case 27:
    glUseProgram(0);
    glDeleteProgram(shaders);
    exit(0);
    break;
  default:
    break;
  }
}

void Reshape(int width, int height)
{
  glViewport(0, 0, width, height);
  aspect = float(height) / float(width);
}

int main(int argc, char ** argv)
{
#ifdef _DEBUG
  glutInitContextFlags(GLUT_DEBUG);
#endif

  //! Connect OpenGL with a Window system
  glutInit(&argc, argv);
  glutInitContextProfile(GLUT_CORE_PROFILE);
  glutInitContextVersion(3, 1);
  glutInitWindowSize(1024, 768);
  glutInitWindowPosition(140, 140);
  glutInitDisplayMode(GLUT_RGBA | GLUT_DOUBLE | GLUT_DEPTH);
  glutCreateWindow("OpenGL Application");

  //! Specify callback functions
  glutDisplayFunc(Display);
  glutReshapeFunc(Reshape);
  glutKeyboardFunc(Keyboard);
  glutMouseFunc(onMouse);
  glutMotionFunc(onMotion);
	
  //! Initialize glew
#ifdef WIN32
  glewInit();
#endif

  // Initialize Opengl context
  Initialize();
  // Endless loop waiting for events
  glutMainLoop();

  return 0;
}





